from ciomaya.lib import scraper_utils
import pymel.core as pm
"""
A scraper to collect paths from Redshift nodes.
"""


ATTRS = {
    "redshift4Maya": {
        "RedshiftBokeh": [
            "dofBokehImage"
        ],
        "RedshiftCameraMap":
        [
            "tex0"],
        "RedshiftDomeLight":
        [
            "tex0",
            "tex1"
        ],
        "RedshiftEnvironment": [
            "tex0",
            "tex1",
            "tex2",
            "tex3",
            "tex4"
        ],

        "RedshiftIESLight":
        [
            "profile"
        ],
        "RedshiftLensDistortion":
        [
            "LDimage"
        ],
        "RedshiftLightGobo":
        [
            "tex0"
        ],
        "RedshiftNormalMap":
        [
            "tex0"
        ],
        "RedshiftOptions": [
            "irradianceCacheFilename",
            "irradiancePointCloudFilename",
            "photonFilename",
            "subsurfaceScatteringFilename"
        ],
        "RedshiftPostEffects": [
            "clrMgmtOcioFilename",
            "lutFilename"
        ],
        "RedshiftProxyMesh":
        [
            "computedFileNamePattern"
        ],
        "RedshiftSprite":
        [
            "tex0"
        ],
        "RedshiftVolumeShape": [
            "computedFileNamePattern"
        ]
    }
}

TOKENS = (r"<UDIM>", r"<f\d?>",r"<Frame>" , r"#+")


def run(_):

    paths = scraper_utils.get_paths(ATTRS)
    paths = scraper_utils.starize_tokens(paths, *TOKENS)
    paths = scraper_utils.expand_env_vars(paths)
    paths = scraper_utils.expand_workspace(paths)
    paths += _scrape_proxies()
    return paths

def _scrape_proxies():
    result = []
    for node in  pm.ls(type="RedshiftProxyMesh"):
        path = node.attr("computedFileNamePattern").get()
        if path:
            for contained_path in pm.mel.eval( 'rsProxy -q -dependencies "{}"'.format(path)) or []:
                result.append({"path":contained_path, "proxy_node": node.name()})
    return result




"""

"""


import importlib
import os
import pymel.core as pm



def has_letters(rhs):
    return rhs.lower().islower()

def get_module_docstring(script):
    try:
        doctring = importlib.import_module(script).__doc__
        if has_letters(doctring):
            return doctring.strip()
    except BaseException:
        pass
    return "No doctring"

def run(node, scripts):
    result = []
    for script in scripts:
        try:
            scraper_module = importlib.import_module(script)
            reload(scraper_module)
        except ImportError:
            pm.warning(
                "Can't load the script '{}' as a Python module.".format(script))
            raise
        except SyntaxError:
            pm.warning(
                "Syntax error in the scraper script: '{}'.".format(script))
            raise
        result += scraper_module.run(node)
    return result




"""
Software data as a singleton.

Software is converted to a PackageTree containing only maya packages and
supported plugins.

Also has the ability to use fixtures for dev purposes.
"""
import re

import pymel.core as pm

def detect_host():
    version_parts = pm.about(iv=True).replace(
        "Autodesk Maya", "").strip().split(".")
    if len(version_parts) < 2:
        version_parts.append("0")
    version_parts[1] = "SP{}".format(version_parts[1])
    return "maya-io {}".format(".".join(version_parts))


def get_plugin_version(plugin):
    try:
        return pm.pluginInfo(plugin, q=True, version=True)
    except RuntimeError:
        return

def detect_mtoa():
    version = get_plugin_version("mtoa")
    if version:
        parts = version.split(".")
        version = ".".join(parts + ["0"] * (4 - len(parts)))
        return "arnold-maya {}".format(version)


def detect_rfm():
    version = get_plugin_version("RenderMan_for_Maya")
    if version:

        parts = filter(None, re.split(r" |\.|@", version))
        version = ".".join(parts + ["0"] * (4 - len(parts)))
        return "renderman-maya {}".format(version)


def detect_vray():
    version = get_plugin_version("vrayformaya")
    if version:  # something loaded
        if version.lower() != "next":
            return version
        try:
            version = pm.fileInfo('vrayBuild', q=True)
        except KeyError:
            return "v-ray-maya 99.99.99.99"

        parts = version.split(" ")[0].split(".")
        version = ".".join(parts + ["0"] * (4 - len(parts)))
        return "v-ray-maya {}".format(version)


def detect_redshift():
    version = get_plugin_version("redshift4maya")
    if version:
        parts = version.split(".")
        version = ".".join(parts + ["0"] * (4 - len(parts)))
        return "redshift-maya {}".format(version)

def detect_yeti():
    version = get_plugin_version("pgYetiMaya")
    if version:
        parts = version.split(".")
        version = ".".join(parts + ["0"] * (4 - len(parts)))
        return "yeti {}".format(version)



"""
Handle the UI for layers

"""
import pymel.core as pm
from ciomaya.lib import const as k
from ciomaya.lib.ae import AEcommon


def create_ui(node_attr):
    """Build static UI.

    Use a customUI because:
    1. TODO: Don't give user the ability to override layers.
    2. TODO: Provide our own custom popup.
    3. TODO: Info list of the layers that will be rendered.
    """
    with AEcommon.ae_template():
        
       
        pm.rowLayout(
                     numberOfColumns=2,
                     columnWidth2=(k.AE_TEXT_WIDTH, 200),
                     columnAttach=((1, "right", 0), (2, "both", 0))
                     )

        pm.text(label="Render Layers")
        pm.attrEnumOptionMenu("layersMenu", label="",  attribute=node_attr, width=200)
        pm.setParent("..")
        populate_ui(node_attr)


def populate_ui(node_attr):
    """Reconfigure action buttons when node changes."""
    widgets = _get_widgets()
    pm.attrEnumOptionMenu(widgets["menu"], edit=True, attribute=node_attr)


def _get_widgets(parent=None):
    if not parent:
        parent = pm.setParent(q=True)

    return {
        "menu":  AEcommon.find_ui("layersMenu", parent),
    }
 
 


import sys

# The template mut be imported before the plugin node.
import maya.api.OpenMaya as om
from ciocore import data as coredata
from ciomaya.lib.ae import AEconductorRenderTemplate
from ciomaya.lib import conductor_menu
from ciomaya.lib.nodes.conductorRender import conductorRender


def maya_useNewAPI():
    pass


def initializePlugin(obj):
    # Use "dev.999 to cause the version to be replaced at build time."
    plugin = om.MFnPlugin(obj, "Conductor", "dev.999", "Any")
    try:
        plugin.registerNode(
            "conductorRender",
            conductorRender.id,
            conductorRender.creator,
            conductorRender.initialize,
            om.MPxNode.kDependNode,
        )
    except:
        sys.stderr.write("Failed to register conductorRender\n")
        raise

    conductor_menu.load()
    
    coredata.init(product="maya-io")


def uninitializePlugin(obj):
    plugin = om.MFnPlugin(obj)

    try:
        plugin.deregisterNode(conductorRender.id)
    except:
        sys.stderr.write("Failed to deregister conductorRender\n")
        raise

    conductor_menu.unload()




import os
import re
import sys

from ciocore.gpath import Path

import maya.api.OpenMaya as om
import pymel.core as pm

PLATFORM = sys.platform
PLACEHOLDER = "CIOPLACEHOLDER"  # Can be anything unique.
ATTR_TOKEN_REGEX = re.compile(
    r"<attr:([0-9a-zA-Z_]+)(?:\s+index:([0-9]+))?(?:\s+default:([0-9a-zA-Z_]+))?>")


def get_paths(attrs):
    """
    Get paths from attributes.

    First get the plugs and iterate. If the leaf level plug is an array, then we
    find out with isArray() and iterate over its elements.

    """
    result = []
    plug_list = _get_plugs(attrs)
    plug_iter = om.MItSelectionList(plug_list)
    while not plug_iter.isDone():
        plug = plug_iter.getPlug()
        plug_iter.next()

        if plug.isArray:
            for index in xrange(plug.numElements()):
                child_plug = plug.elementByPhysicalIndex(index)
                value = _get_value(child_plug)
                if value:
                    result.append(_get_value(child_plug))
        else:
            value = _get_value(plug)
            if value:
                result.append(_get_value(plug))
    return result


def _get_plugs(attrs):
    """
    Return a SelectionList that contains the actual existing plugs.

    When we get a node's plug from the att name, it may be a child of a compound
    array plug (or nested several levels). In this case the plug name will be of
    the form node_name.parentPlug[-1].childPlug This (-1) is a nonexistent plug.
    In order to get the actual plug elements, if they exist, we use a
    SelectionList. We add a wildcard name to the selectionlist to get the real
    plugs it contains.

    The [-1] indicator is only used for array parent plugs, not leaf level array
    plugs.

    [
        "someGrouping": {
            "someNodeType": [
                "topLevelAttributeName",
                "nestedAttributeName",
                "arrayTypeAttributeName",
                "childOfNestedArrayTypeAttributeName"
            ],
            ...
        },
        ...
    ]
    """
    all_node_types = pm.allNodeTypes()
    selection_list = om.MSelectionList()
    for section in attrs:
        for nodetype in attrs[section]:
            if nodetype in all_node_types:
                for node in pm.ls(type=nodetype):
                    for attr in attrs[section][nodetype]:
                        try:
                            selection_list.add(
                                node.attr(attr).name().replace("[-1]", "[*]"))
                        except (RuntimeError, pm.MayaAttributeError):
                            pass
    return selection_list


def _get_value(plug):
    value = plug.asString()
    if value:
        return {"path": value, "plug": plug.name()}


def starize_tokens(paths, *tokens):
    """
    Replace any of the given tokens with a '*'
    """
    token_rx = re.compile("|".join(tokens), re.IGNORECASE)
    for p in paths:
        p["path"] = token_rx.sub("*", p["path"])
    return paths


def expand_env_vars(paths):
    """Use Path to resolve env vars"""
    for i  in range(len(paths)):
        try:
            paths[i]["path"] = Path(paths[i]["path"]).posix_path()
        except TypeError:
            paths[i] = Path(paths[i]).posix_path()
    return paths


def expand_workspace(paths):
    """
    Expand in a non-platform-dependent way.

    Paths are either a list of strings, a list of objects with a "path" key, or
    a mix.

    Sometimes on a Mac/Linux, the user will have a scene with Windows paths in
    it. That's a mistake of course, but we want to show them the mistake. If we
    just expandName, then those windows paths will be treated as relative. The
    expandName function prepends the workspace, and to add insult to injury, it
    deletes everything after the drive letter. You end up with
    /Volumes/blah/blah/C So we check that its absolute before running through
    expandName and we leave it in tact with its drive letter so the user can
    identify it during validation.
    """
    ws = pm.Workspace()
    for i, p in enumerate(paths):
        try:
            if Path(p["path"]).relative:
                paths[i]["path"] = ws.expandName(p["path"])
        except TypeError:
            if Path(p).relative:
                paths[i] = ws.expandName(p)
    return paths


def extend_with_tx_paths(paths):
    """
    Add the tx sister for image files

    Use glob notation around one letter in the extension (.t[x]) because when
    paths are finally resolved, the list is expanded by globbing.

    As TX files are not critical, we don't want to block the submission if they
    don't exist. Glob will ultimately expand to no files if the file does not
    exist.
    """

    image_ext_regex = re.compile(
        r"^\.(jpg|jpeg|gif|iff|psd|png|pic|tga|tif|tiff|bmp|hdr)$", re.IGNORECASE)
    txpaths = []
    for p in paths:
        root, ext = os.path.splitext(p["path"])
        if image_ext_regex.match(ext):
            txpath = "{}.t[x]".format(root)
            txpaths.append({"plug": p["plug"], "path": txpath})
    return paths+txpaths


def resolve_to_sequence(template, sequence):
    """
    Replace all the popular frame placeholders with a consistent template
    expression that specifies the padding
    """

    orig_template = template
    # replace 4 hashes with  {frame:04d}
    template = re.compile(r"(#+)",re.IGNORECASE).sub(
        lambda match: "{{frame:0{:d}d}}".format(len(match.group(1))), 
        template)

    # replace $F4 with {frame:04d}
    template = re.compile(r"\$F(\d?)",re.IGNORECASE).sub( 
         lambda match: "{{frame:0{:d}d}}".format(int(match.group(1) or "1"))  ,
         template)

    # replace <f4> with {frame:04d}
    template = re.compile(r"<F(\d?)>",re.IGNORECASE).sub( 
         lambda match: "{{frame:0{:d}d}}".format(int(match.group(1) or "1"))  ,
         template)

    # replace <frame> with {frame}
    template = re.compile(r"<frame>",re.IGNORECASE).sub("{frame}", template)

    if orig_template == template:
        return [orig_template]

    return sequence.expand_format(template)



def extract_attr_token(filename):
    """   
    Find <attr: token in  paths specified like so:

    /Volumes/xtr/gd/standin_fixture//sourceimages/<attr:floormap default:alcazar>.jpg

    # https://docs.arnoldrenderer.com/pages/viewpage.action?pageId=40110953

    # REGEX TESTER
    # https://regex101.com/r/eFp4RT/1/
    # """ 
    match = ATTR_TOKEN_REGEX.search(filename)
    if not match:
        return

    # cant do indices
    if match.group(2) is not None:
        return

    attr_name = match.group(1)
    default_val = match.group(3)
    template = ATTR_TOKEN_REGEX.sub(PLACEHOLDER, filename)

    return template, attr_name, default_val




"""
Handle the UI for output:

"""

import pymel.core as pm
from ciomaya.lib.ae import AEcommon

import json

def create_ui(node_attr):
    """Build static UI."""
    with AEcommon.ae_template():
        pm.scrollField("outputScrollField", height=250, editable=False, wordWrap=False)
        populate_ui(node_attr)

def populate_ui(node_attr):
    """Reconfigure action buttons when node changes."""
    attr = pm.Attribute(node_attr)
    field = AEcommon.find_ui("outputScrollField", pm.setParent(q=True))
    _on_output_change(attr, field)
    pm.scriptJob(attributeChange=(
        attr, pm.Callback( _on_output_change, attr, field)), parent=field, replacePrevious=True)


def _on_output_change(attr, field):
    data = attr.get()
    text = "Not valid"
    if data:
        try:
            obj =  json.loads(data)
            text = json.dumps(obj, sort_keys=True, indent=4)
        except BaseException:
            pass
    pm.scrollField(field, edit=True, text=text)
 




import pymel.core as pm
from contextlib import contextmanager

@contextmanager
def layer_context(layer):
    """Perform some ops in the context of the given render layer"""
    current_layer = pm.editRenderLayerGlobals(
        q=True, currentRenderLayer=True)
    pm.editRenderLayerGlobals(currentRenderLayer=layer)
    yield
    pm.editRenderLayerGlobals(currentRenderLayer=current_layer)


def get_layer_name(layer):
    layer = pm.PyNode(layer)
    name = layer.name()
    if name == "defaultRenderLayer" or name == "masterLayer":
        return "masterLayer"
    conns = layer.attr("message").connections(
        d=True, s=False, et=True, t="renderSetupLayer")
    if not conns:
        return name
    return conns[0].name()

def get_current_layer_name():
    return get_layer_name(pm.editRenderLayerGlobals(q=True, currentRenderLayer=True))


def get_renderable_legacy_layers():
    return [l for l in pm.PyNode("renderLayerManager").attr(
        "renderLayerId").connections() if l.attr("renderable").get()]



"""
Handle the UI for emails:
"""

import pymel.core as pm
from ciomaya.lib import const as k

from ciomaya.lib.ae import AEcommon


def create_ui(node_attr):
    """Build static UI."""
    with AEcommon.ae_template():

        pm.rowLayout(
            numberOfColumns=2,
            columnWidth2=(k.AE_TEXT_WIDTH, 210),
            columnAttach=((1, "right", 0), (2, "both", 0)),
        )
        pm.text(label="")
        pm.button("emailsAddButton", label="Add Email Address", height=24)
        #  command=pm.Callback(on_add_btn, node)
        pm.setParent("..")  # out rowLayout
        pm.separator()

        pm.columnLayout("emailsColumn", adj=True)

        pm.setParent("..")  # out columnLayout

        populate_ui(node_attr)


def populate_ui(node_attr):
    """Reconfigure action buttons when node changes."""

    attr = pm.Attribute(node_attr)
    widgets = _get_widgets()

    pm.button(widgets["button"], edit=True,
              command=pm.Callback(_on_add_btn, attr))

    for widget in pm.columnLayout(widgets["column"], q=True, childArray=True) or []:
        pm.deleteUI(widget)

    pm.setParent(widgets["column"])
    for i, attr_element in enumerate(attr):

        address_att = attr_element.attr("emailAddress")
        active_att = attr_element.attr("emailAddressActive")

        label_ctl = pm.text(label="Email: {:d}".format(i + 1))

        address_ctl = pm.textField()
        active_ctl = pm.checkBox(label="")

        del_ctl = pm.symbolButton(
            image="smallTrash.xpm", command=pm.Callback(_on_remove_email, attr_element)
        )

        _form_layout(label_ctl, address_ctl, active_ctl, del_ctl)

        pm.connectControl(active_ctl, active_att)
        pm.connectControl(address_ctl, address_att)

        _setup_script_jobs(active_att, address_ctl)

        _active_cb_changed(active_att, address_ctl)

        pm.setParent(widgets["column"])


def _get_widgets(parent=None):
    if not parent:
        parent = pm.setParent(q=True)
    return {
        "button":  AEcommon.find_ui("emailsAddButton", parent),
        "column":  AEcommon.find_ui("emailsColumn", parent)
    }


def _setup_script_jobs(active_att, address_ctl):

    pm.scriptJob(
        attributeChange=(
            active_att,
            pm.Callback(_active_cb_changed, active_att, address_ctl),
        ),
        parent=address_ctl,
        replacePrevious=True
    )


def _on_add_btn(attr):
    AEcommon.next_available_element_plug(attr).attr("emailAddress").set("")


def _active_cb_changed(att, ctl):
    pm.textField(ctl, edit=True, enable=att.get())


def _on_remove_email(attr):
    pm.removeMultiInstance(attr, b=True)


def _form_layout(*widgets):
    # There must be 4 widgets.
    form = pm.formLayout(nd=100)
    for widget in widgets:
        pm.control(widget, edit=True, parent=form)

    form.attachForm(widgets[0], "left", 2)
    form.attachNone(widgets[0], "right")
    form.attachForm(widgets[0], "top", 2)
    form.attachForm(widgets[0], "bottom", 2)

    form.attachControl(widgets[1], "left", 2, widgets[0])
    form.attachControl(widgets[1], "right", 2, widgets[2])
    form.attachForm(widgets[1], "top", 2)
    form.attachForm(widgets[1], "bottom", 2)

    form.attachNone(widgets[2], "left")
    form.attachControl(widgets[2], "right", 2, widgets[3])
    form.attachForm(widgets[2], "top", 2)
    form.attachForm(widgets[2], "bottom", 2)

    form.attachNone(widgets[3], "left")
    form.attachForm(widgets[3], "right", 2)
    form.attachForm(widgets[3], "top", 2)
    form.attachForm(widgets[3], "bottom", 2)

    return form



from ciomaya.lib import scraper_utils

"""
A scraper to collect paths for Vray.
"""


ATTRS = {
    "vray": {
        "VRayMesh": [
            "fileName"
        ],
        "VRaySettingsNode": [
            "ifile",
            "fnm"
        ],
        "VRayVolumeGrid": [
            "inFile",
            "inPath",
        ],
        "VRayScene": [
            "FilePath"
        ]
    }
}

# See https://docs.chaosgroup.com/display/VMAYA/File+Names+for+Bitmap+Textures
TOKENS = (r"<UDIM>", r"<UVTILE>", r"<frameNum>",
          r"\$\d*U.*\$\d*V", r"u<U>_v<V>", r"u<U>_v<V>_<f>", r"#+")


def run(_):

    paths = scraper_utils.get_paths(ATTRS)
    paths = scraper_utils.starize_tokens(paths, *TOKENS)
    paths = scraper_utils.expand_workspace(paths)
    return paths



"""
A basic scraper to collect paths from Maya nodes.
"""
import pymel.core as pm
import re
from ciomaya.lib import scraper_utils

 
ATTRS = {
    "AbcImport": {
        "AlembicNode": ["abc_File"]
    },
    "MayaBuiltin": {
        "file": ["computedFileTextureNamePattern"],
        "cacheFile": ["cachePath"],
        "gpuCache": ["cacheFileName"]
    }
}

TOKENS = (
    r"#+"  # image.####.exr - hash
    , r"_MAPID_"  # image_MAPID_.exr - mapid
    , r"%0\d+d"  # image.%04d.exr - percent
    , r"<UDIM>"  # image.<UDIM>.exr - udim_mari
    , r"<UVTILE>"  # image.<UVTILE>.exr - udim_mudbox
    , r"\$\d*U.*\$\d*V"  # image.u$2U_v$2V.exr or image.u$Uv$V.exr, etc - udim_vray
    , r"u<U>_v<V>"  # image.u<U>_v<V>.exr - udim_zbrush
    , r"u<U>_v<V>_<f>"  # image.u<U>_v<V>_<f>.exr - udim_zbrush_f
    , r"<f\d?>"  # image.<f>.exr or image.<f4>.exr - frame_seq
    , r"<FrameNum>"  # image.<FrameNum>.ext - Redshift
    , r"<Frame>"  # image.<Frame>.ext - Redshift
)


 
def run(_):
    """
    Find paths in Maya nodes.

    Since Maya nodes may be rendered by other renderers, and those renderers may
    support a variety of tokens, we include all tokens here. 

    Example: Renderman uses _MAPID_, but it could be found in a Maya file node. Leaky abstraction?
    Perhaps the better solution is to include the file node in the renderman scraper?

    Image file textures may have a sibling ".tx file".
    """

    paths = scraper_utils.get_paths(ATTRS)
    paths = scraper_utils.starize_tokens(paths, *TOKENS)
    paths = scraper_utils.expand_workspace(paths)
    paths = scraper_utils.extend_with_tx_paths(paths)
    paths.extend(_get_scene_files())

    return paths



def _get_scene_files():
    result = []
    depth_regex = re.compile(r".*{(\d+)\}$")
    for p in pm.listReferences(recursive=True):
        path = unicode(p)
        match = depth_regex.match(path)
        depth = 1
        if match:
            path = re.sub("{(\d+)\}$", "", path)
            depth = int(match.group(1))+1

        result.append({"path":path, "depth":depth})

    scene_name =  unicode(pm.sceneName())
    if scene_name:
        result.append({"path": scene_name, "depth":0})
    return result




from ciomaya.lib import scraper_utils

"""
A scraper to collect paths from Renderman nodes.
"""


ATTRS = {
    "Renderman_for_Maya": {
        "PxrBump": [
            "filename"
        ],
        "PxrCookieLightFilter": [
            "map"
        ],
        "PxrDiskLight": [
            "iesProfile"
        ],

        "PxrDomeLight": [
            "lightColorMap"
        ],

        "PxrGobo": [
            "map"
        ],
        "PxrGoboLightFilter": [
            "map"
        ],

        "PxrLayeredTexture": [
            "maskTexture",
            "filename"
        ],

        "PxrMultiTexture": [
            "filename0",
            "filename1",
            "filename2",
            "filename3",
            "filename4",
            "filename5",
            "filename6",
            "filename7",
            "filename8",
            "filename9",
        ],
        "PxrNormalMap": [
            "filename"
        ],
        "PxrOSL": [
            "shadername"
        ],

        "PxrProjectionLayer": [
            "channelsFilenames",
            "filename"
        ],
        "PxrPtexture": [
            "filename"
        ],

        "PxrRectLight": [
            "lightColorMap",
            "iesProfile"
        ],

        "PxrSphereLight": [
            "iesProfile"
        ],
        "PxrStdAreaLight": [
            "profileMap",
            "rman__EmissionMap",
            "iesProfile",
            "barnDoorMap"
        ],

        "PxrStdEnvMapLight": [
            "rman__EnvMap"
        ],
        "PxrTexture": [
            "filename"
        ],
        "PxrVisualizer": [
            "matCap"
        ],
        "RenderManArchive": [
            "filename"
        ],
        "rmanImageFile": [
            "File"
        ],
        "rmanTexture3d": [
            "File"
        ],
        "RMSAreaLight": [
            "mapname"
        ],
        "RMSCausticLight": [
            "causticPhotonMap"
        ],
        "RMSEnvLight": [
            "rman__EnvMap"
        ],
        "RMSGPSurface": [
            "SpecularMapB",
            "SpecularMap",
            "RoughnessMap",
            "MaskMap",
            "SurfaceMap",
            "DisplacementMap"
        ],
        "RMSGeoAreaLight": [
            "profilemap",
            "iesprofile",
            "lightcolormap",
            "barnDoorMap"
        ],
        "RMSGeoLightBlocker": [
            "Map"
        ],
        "RMSGlass": [
            "roughnessMap",
            "surfaceMap",
            "specularMap",
            "displacementMap"
        ],
        "RMSLightBlocker": [
            "Map"
        ],
        "RMSMatte": [
            "SurfaceMap",
            "MaskMap",
            "DisplacementMap"
        ],

        "RMSOcean": [
            "roughnessMap",
            "surfaceMap",
            "specularMap",
            "displacementMap"
        ]
    }
}

# See https://rmanwiki.pixar.com/display/RFM22/String+tokens+in+RfM
TOKENS = (r"_MAPID_", r"<udim>", r"<frame>", r"<f\d?>", r"<aov>", r"#+")


def run(_):

    paths = scraper_utils.get_paths(ATTRS)
    paths = scraper_utils.starize_tokens(paths, *TOKENS)
    paths = scraper_utils.expand_workspace(paths)
    paths = scraper_utils.extend_with_tx_paths(paths)
    return paths



"""
Handle the UI for actions:

"""
import pymel.core as pm
from ciocore import data as coredata

from ciomaya.lib.ae import AEcommon
from ciomaya.lib import const as k
from ciomaya.lib import submit
from ciomaya.lib import validation
from ciocore.validator import ValidationError
from ciomaya.lib import asset_cache


def create_ui(node_attr):
    """Build row for action buttons."""
    with AEcommon.ae_template():
        but_width = k.AE_TOTAL_WIDTH / 6
        top = pm.frameLayout(label="Shelf", cll=False, cl=False)
        pm.rowLayout(
            numberOfColumns=3,
            cw3=(but_width, but_width, but_width),
            columnAttach3=("left", "left", "left"),
        )

        pm.iconTextButton(
            "connectButton",
            label="Reconnect",
            ann="Reconnect to Conductor and fetch account data",
            w=but_width,
            style="iconAndTextVertical",
            en=True,
            image1="ConductorConnect_30x30.png",
        )

        pm.iconTextButton(
            "submitButton",
            label="Submit",
            ann="Submit Job",
            w=but_width,
            style="iconAndTextVertical",
            en=False,
            image1="ConductorSubmit_30x30.png",
        )

        pm.iconTextButton(
            "validateButton",
            label="Validate",
            ann="Validate Only",
            w=but_width,
            style="iconAndTextVertical",
            image1="ConductorShow_30x30.png",
        )

        pm.setParent(top)
        pm.setParent("..")

        populate_ui(node_attr)


def populate_ui(node_attr):
    """Reconfigure action buttons when node changes."""
    widgets = _get_widgets()
    node = pm.Attribute(node_attr).node()
    pm.iconTextButton(
        widgets["connect"], edit=True, command=pm.Callback(
            _on_connect, widgets, node, force=True)
    )
    pm.iconTextButton(
        widgets["submit"], edit=True, en=False, command=pm.Callback(
            on_submit, node)
    )

    pm.iconTextButton(
        widgets["validate"], edit=True,  command=pm.Callback(
            on_validate, node)
    )

    pm.evalDeferred(pm.Callback(_on_connect, widgets, node))


def _on_connect(widgets, node, force=False):

    use_fixtures = node.attr("useFixtures").get()
    print "use_fixtures is currently", use_fixtures
    fixtures_dir = node.attr(
        "fixturesDirectory").get() if use_fixtures else None
    print "use_fixtures is currently", use_fixtures
    coredata.set_fixtures_dir(fixtures_dir)

    try:
        coredata.data(force=force)
    except BaseException as ex:
        pm.displayError(str(ex))
        pm.displayWarning(
            "Try again after deleting your credentials file (~/.config/conductor/credentials)")
    set_enabled_state(widgets)

    if force:
        # If the Reconnect button was clicked, then also force refresh in order to tell the 
        # inst_types, software, and projects rebuild themselves based on new data.
        pm.mel.openAEWindow()


def _get_widgets(parent=None):
    """Widgets are children of the named top layout we constructed.

    If no parent given, then twe must be "in" the parent
    that contains the expected widgets.
    """
    if not parent:
        parent = pm.setParent(q=True)

    return {
        "submit":  AEcommon.find_ui("submitButton", parent),
        "connect":  AEcommon.find_ui("connectButton", parent),
        "validate":  AEcommon.find_ui("validateButton", parent)

    }


def set_enabled_state(widgets):
    can_submit = coredata.valid()
    pm.iconTextButton(widgets["submit"], edit=True, en=can_submit)


def on_submit(node):
    asset_cache.clear()
    submit.submit(node)


def on_validate(node):
    asset_cache.clear()
    try:
        validation.run(node, dry_run=True)
    except ValidationError as ex:
        pm.displayWarning(str(ex))



"""
Entry point for the conductorRender Attribute Editor UI.

Generally, there's one attributeEditor instance for the conductorRender node type. When different nodes of that
type are selected, the same AE is shown and its contents replaced to reflect the current node.

AE templates build interfaces to each attribute in one of 2 ways:
Simple: e.g. self.addControl("preemptible")
Custom: e.g. self.callCustom(new_func, replace_func, "preemptible")

new_func builds the UI the first time the template is run. replace_func
reconfigures the template based on the current node. The purpose is efficiency -
so that the template doesn't need to rebuild everything everytime.

We call out to separate files (classes) to build each section if it uses a custom UI.
e.g. AEsoftware() is a class containing a new_func and a replace_func to manage the software
section.

"""
import pymel.core as pm
from ciomaya.lib import const as k
from ciomaya.lib.ae import (AEactions, AEdestination,
                            AEemails, AEenvironment, AEextraAssets,
                            AEframes, AEinstanceType, AElayers,
                            AEmetadata, AEoutput, AEproject, AEscrapers,
                            AEsoftware, AEtaskTemplate)


class AEconductorRenderTemplate(pm.ui.AETemplate):
    def __init__(self, node_name):
        """
        Define the high level arrangement of AE sections.
        """
        pm.ui.AETemplate.__init__(self, node_name)

        self.beginScrollLayout()

        self.callCustom(AEactions.create_ui, AEactions.populate_ui, "title")

        self.beginLayout("General Attributes", collapse=False)
        self.addControl("title")

        self.callCustom(AEproject.create_ui,
                        AEproject.populate_ui, "projectName")

        self.callCustom(AElayers.create_ui,
                        AElayers.populate_ui, "renderLayers")

        self.addSeparator()

        self.callCustom(AEinstanceType.create_ui,
                        AEinstanceType.populate_ui, "instanceTypeName")

        self.addControl("preemptible")
        self.addSeparator()

        self.callCustom(AEdestination.create_ui,
                        AEdestination.populate_ui, "destinationDirectory")

        self.endLayout()

        self.beginLayout("Software", collapse=False)
        self.callCustom(AEsoftware.create_ui,
                        AEsoftware.populate_ui, "hostSoftware")
        self.endLayout()

        self.beginLayout("Frame range")
        self.addControl("chunkSize")

        self.callCustom(AEframes.create_ui,
                        AEframes.populate_ui, "customRange")
        self.callCustom(AEframes.create_ui,
                        AEframes.populate_ui, "scoutFrames")

        self.endLayout()

        self.beginLayout("Info", collapse=False)
        self.addControl("frameSpec")
        self.addControl("scoutSpec")
        self.addSeparator()
        self.addControl("frameCount")
        self.addControl("taskCount")
        self.addControl("scoutTaskCount")
        self.endLayout()

        self.beginLayout("Assets", collapse=False)

        self.addControl("useUploadDaemon",
                        changeCommand=self.update_use_upload_daemon)
        self.addControl("uploadOnly")

        self.beginLayout("Asset Scrapers")
        self.callCustom(AEscrapers.create_ui,
                        AEscrapers.populate_ui, "assetScrapers")
        self.endLayout()

        self.beginLayout("Extra Assets")
        self.callCustom(AEextraAssets.create_ui,
                        AEextraAssets.populate_ui, "extraAssets")
        self.endLayout()

        self.endLayout()

        self.beginLayout("Notifications")
        self.callCustom(AEemails.create_ui,
                        AEemails.populate_ui, "emailAddresses")
        self.endLayout()

        self.beginLayout("Task Command")
        # self.addControl("taskTemplate")
        self.callCustom(AEtaskTemplate.create_ui,
                        AEtaskTemplate.populate_ui, "taskTemplate")

        self.endLayout()

        self.beginLayout("Metadata")

        self.callCustom(AEmetadata.create_ui,
                        AEmetadata.populate_ui, "metadata")
        self.endLayout()

        self.beginLayout("Extra Environment")
        self.callCustom(AEenvironment.create_ui,
                        AEenvironment.populate_ui, "extraEnvironment")
        self.endLayout()

        self.beginLayout("Automatic Retries")
        self.addControl("retriesWhenPreempted")
        self.addControl("retriesWhenFailed")
        self.endLayout()

        self.beginLayout("Submission Preview")
        self.addControl("doScrape", label="Display Scraped Assets")
        self.addControl("taskLimit", label="Display Tasks")
        self.callCustom(AEoutput.create_ui,
                        AEoutput.populate_ui, "output")

        self.endLayout()

        self.beginLayout("Autosave")
        self.addControl("autosave")
        self.addControl("autosaveTemplate")
        self.addControl("cleanupAutosave")
        self.endLayout()

        self.beginLayout("Location")
        self.addControl("locationTag")
        self.endLayout()

        self.beginLayout("Developer")
        self.addControl("showTracebacks")
        self.addControl("fixturesDirectory")
        self.addControl("useFixtures",
                        changeCommand=self.update_use_fixtures)
        self.endLayout()

        self.addExtraControls()

        for att in k.SUPPRESS_EXTRA_ATTS:
            self.suppress(att)

        self.endScrollLayout()

    def update_use_upload_daemon(self, nodeName):
        use_daemon = pm.PyNode(nodeName).attr("useUploadDaemon").get()
        self.dimControl(nodeName, "cleanupAutosave", use_daemon)

    def update_use_fixtures(self, nodeName):
        dim = not pm.PyNode(nodeName).attr("useFixtures").get() 


        self.dimControl(nodeName, "fixturesDirectory", dim)



"""
Handle the UI for environment:
"""

import pymel.core as pm
from ciomaya.lib import const as k
from ciomaya.lib.ae import AEcommon


def create_ui(node_attr):
    """Build static UI."""
    with AEcommon.ae_template():
 
        pm.rowLayout(
            width=290,
            numberOfColumns=2,
            columnWidth2=(k.AE_TEXT_WIDTH, 210),
            columnAttach=((1, "right", 0), (2, "both", 0)),
        )

        pm.text(label="")
        pm.button("metadataAddButton",label="Add Metadata Entry", height=24 )
        pm.setParent("..")  # out rowLayout
        pm.separator()

        _form_layout(pm.text(label="Key", align="left", width=k.AE_TEXT_WIDTH),
                     pm.text(label="Value", align="left"),
                     pm.text(label="", width=36)
                     )
        
        pm.setParent("..")  # out _form_layout

        pm.columnLayout("metadataColumn", adj=True)

        pm.setParent("..")  # out columnLayout

        populate_ui(node_attr)




def populate_ui(node_attr):
    """Reconfigure action buttons when node changes."""

    attr = pm.Attribute(node_attr)
    widgets = _get_widgets()

    pm.button(widgets["button"], edit=True,
              command=pm.Callback(_add_entry, attr))

    for widget in pm.columnLayout(widgets["column"], q=True, childArray=True) or []:
        pm.deleteUI(widget)

    pm.setParent(widgets["column"])
    for attr_element in attr:

        key_att = attr_element.attr("metadataKey")
        val_att = attr_element.attr("metadataValue")

        key_ctl = pm.textField(placeholderText="METADATA_KEY", width=k.AE_TEXT_WIDTH)
        val_ctl = pm.textField(placeholderText="METADATA_VALUE")
        del_ctl = pm.symbolButton(
            image="smallTrash.xpm", width=36, command=pm.Callback(_remove_entry, attr_element)
        )

        _form_layout(key_ctl, val_ctl, del_ctl)
        pm.setParent("..")  # out _form_layout


        pm.connectControl(key_ctl, key_att)
        pm.connectControl(val_ctl, val_att)

        pm.setParent(widgets["column"])

def _get_widgets(parent=None):
    if not parent:
        parent = pm.setParent(q=True)
    return {
        "button":  AEcommon.find_ui("metadataAddButton", parent),
        "column":  AEcommon.find_ui("metadataColumn", parent)
    }



def _add_entry(attr):
    AEcommon.next_available_element_plug(attr).attr("metadataKey").set("")
 
def _remove_entry(attribute):
    pm.removeMultiInstance(attribute, b=True)


def _form_layout(*widgets):
    # There must be 3 widgets.
    form = pm.formLayout(nd=100)
    for widget in widgets:
        pm.control(widget, edit=True, parent=form)

    form.attachForm(widgets[0], "left", 2)
    form.attachNone(widgets[0], "right")
    form.attachForm(widgets[0], "top", 2)
    form.attachForm(widgets[0], "bottom", 2)

    form.attachControl(widgets[1], "left", 2, widgets[0])
    form.attachControl(widgets[1], "right", 2, widgets[2])
    form.attachForm(widgets[1], "top", 2)
    form.attachForm(widgets[1], "bottom", 2)

    form.attachNone(widgets[2], "left")
    form.attachForm(widgets[2], "right", 2)
    form.attachForm(widgets[2], "top", 2)
    form.attachForm(widgets[2], "bottom", 2)

    return form



import glob
import imp
import importlib
import json
import os
import re
import pymel.core as pm
# import maya.cmds as cmds
from ciocore import data as coredata
from ciocore.expander import Expander
from ciocore.gpath import Path
from ciocore.gpath_list import GLOBBABLE_REGEX, PathList
from ciocore.package_environment import PackageEnvironment
from ciocore.sequence import Sequence

import maya.api.OpenMaya as om

from ciomaya.lib import const as k
from ciomaya.lib import layer_utils  # , software


def maya_useNewAPI():
    pass


class conductorRender(om.MPxNode):
    # pass

    # static attributes
    aTitle = None

    aChunkSize = None
    aUseCustomRange = None
    aCustomRange = None
    aStartFrame = None
    aEndFrame = None
    aByFrame = None
    aAnimation = None
    aUseScoutFrames = None
    aScoutFrames = None

    aTaskTemplate = None

    aInstanceTypeName = None
    aPreemptible = None
    aProjectName = None
    aRenderLayers = None
    aCurrentRenderLayer = None

    aHostSoftware = None
    aPluginSoftware = None
    aExtraAssets = None

    aAssetScraperName = None
    aAssetScraperActive = None
    aAssetScrapers = None

    aExtraEnvironment = None
    aExtraEnvironmentKey = None
    aExtraEnvironmentValue = None
    aExtraEnvironmentExclusive = None

    aMetadata = None
    aMetadataKey = None
    aMetadataValue = None

    aUploadOnly = None
    aUseUploadDaemon = None

    aEmailAddresses = None
    aEmailAddress = None
    aEmailAddressActive = None

    aRetriesWhenPreempted = None
    aRetriesWhenFailed = None

    aTaskLimit = None
    aDoScrape = None

    aFrameCount = None
    aTaskCount = None
    aScoutTaskCount = None
    aAssetCount = None
    aAssetsSize = None
    aFrameSpec = None
    aScoutSpec = None
    aCurrentRenderer = None
    aCurrentTime = None

    aDestinationDirectory = None
    aLocationTag = None

    aAutosave = None
    aAutosaveTemplate = None
    aCleanupAutosave = None

    aShowTracebacks = None
    aFixturesDirectory = None
    aUseFixtures = None

    aOutput = None

    id = om.MTypeId(0x880500)

    @staticmethod
    def creator():
        return conductorRender()

    @classmethod
    def initialize(cls):
        cls.make_title_att()
        cls.make_frames_atts()
        cls.make_instance_type_att()
        cls.make_project_name_att()
        cls.make_layer_atts()
        cls.make_software_att()
        cls.make_assets_atts()
        cls.make_environment_atts()
        cls.make_task_atts()
        cls.make_upload_flag_atts()
        cls.make_notification_atts()
        cls.make_metadata_atts()
        cls.make_retries_atts()

        cls.make_hidden_atts()
        cls.make_info_atts()

        cls.make_autosave_atts()
        cls.make_developer_atts()
        cls.make_misc_atts()

        cls.make_output_att()

        cls.setup_attribute_affects()

    @staticmethod
    def _make_output_int_att(longname, shortname):
        nAttr = om.MFnNumericAttribute()
        att = nAttr.create(longname, shortname, om.MFnNumericData.kInt)
        nAttr.storable = False
        nAttr.writable = False
        nAttr.readable = True
        om.MPxNode.addAttribute(att)
        return att

    @classmethod
    def make_info_atts(cls):

        cls.aFrameCount = cls._make_output_int_att("frameCount", "frc")
        cls.aTaskCount = cls._make_output_int_att("taskCount", "tsc")
        cls.aScoutTaskCount = cls._make_output_int_att("scoutTaskCount", "stc")
        cls.aAssetCount = cls._make_output_int_att("assetCount", "asc")
        cls.aAssetsSize = cls._make_output_int_att("assetsSize", "asz")

        tAttr = om.MFnTypedAttribute()
        cls.aFrameSpec = tAttr.create("frameSpec", "fms", om.MFnData.kString)
        tAttr.storable = False
        tAttr.writable = False
        tAttr.readable = True
        om.MPxNode.addAttribute(cls.aFrameSpec)

        cls.aScoutSpec = tAttr.create("scoutSpec", "scs", om.MFnData.kString)
        tAttr.storable = False
        tAttr.writable = False
        tAttr.readable = True
        om.MPxNode.addAttribute(cls.aScoutSpec)

    @classmethod
    def make_title_att(cls):
        tAttr = om.MFnTypedAttribute()
        cls.aTitle = tAttr.create("title", "ttl", om.MFnData.kString)
        tAttr.hidden = False
        tAttr.storable = True
        tAttr.readable = True
        tAttr.writable = True
        om.MPxNode.addAttribute(cls.aTitle)

    @classmethod
    def make_instance_type_att(cls):
        tAttr = om.MFnTypedAttribute()
        nAttr = om.MFnNumericAttribute()
        cls.aInstanceTypeName = tAttr.create(
            "instanceTypeName", "itn", om.MFnData.kString
        )
        tAttr.hidden = False
        tAttr.storable = True
        tAttr.readable = True
        tAttr.writable = True
        om.MPxNode.addAttribute(cls.aInstanceTypeName)

        cls.aPreemptible = nAttr.create(
            "preemptible", "prm", om.MFnNumericData.kBoolean, True
        )
        nAttr.keyable = False
        nAttr.storable = True
        nAttr.readable = True
        nAttr.writable = True
        om.MPxNode.addAttribute(cls.aPreemptible)

    @classmethod
    def make_project_name_att(cls):
        tAttr = om.MFnTypedAttribute()
        cls.aProjectName = tAttr.create(
            "projectName", "prn", om.MFnData.kString
        )
        tAttr.hidden = False
        tAttr.storable = True
        tAttr.readable = True
        tAttr.writable = True
        om.MPxNode.addAttribute(cls.aProjectName)

    @classmethod
    def make_layer_atts(cls):
        eAttr = om.MFnEnumAttribute()
        nAttr = om.MFnNumericAttribute()
        tAttr = om.MFnTypedAttribute()

        cls.aRenderLayers = eAttr.create(
            "renderLayers", "rl", k.CURRENT_LAYER
        )
        eAttr.addField("current", k.CURRENT_LAYER)
        eAttr.addField("renderable", k.RENDERABLE_LAYERS)
        eAttr.hidden = False
        eAttr.keyable = True
        eAttr.storable = True
        om.MPxNode.addAttribute(cls.aRenderLayers)

        cls.aCurrentRenderLayer = nAttr.create(
            "currentRenderLayer", "crl", om.MFnNumericData.kInt
        )
        nAttr.hidden = False
        nAttr.writable = True
        nAttr.keyable = True
        nAttr.storable = False
        om.MPxNode.addAttribute(cls.aCurrentRenderLayer)

        cls.aCurrentRenderer = tAttr.create(
            "currentRenderer", "cren", om.MFnData.kString
        )
        tAttr.hidden = True
        tAttr.storable = True
        tAttr.readable = True
        tAttr.writable = True
        om.MPxNode.addAttribute(cls.aCurrentRenderer)

    @classmethod
    def make_software_att(cls):
        tAttr = om.MFnTypedAttribute()
        cls.aHostSoftware = tAttr.create(
            "hostSoftware", "hsw", om.MFnData.kString
        )
        tAttr.hidden = False
        tAttr.writable = True
        om.MPxNode.addAttribute(cls.aHostSoftware)

        cls.aPluginSoftware = tAttr.create(
            "pluginSoftware", "psw", om.MFnData.kString
        )
        tAttr.array = True
        tAttr.hidden = False
        tAttr.writable = True
        om.MPxNode.addAttribute(cls.aPluginSoftware)

    @classmethod
    def make_assets_atts(cls):
        cAttr = om.MFnCompoundAttribute()
        tAttr = om.MFnTypedAttribute()
        nAttr = om.MFnNumericAttribute()

        cls.aAssetScraperName = tAttr.create(
            "assetScraperName", "asn", om.MFnData.kString
        )

        cls.aAssetScraperActive = nAttr.create(
            "assetScraperActive", "asa", om.MFnNumericData.kBoolean, True
        )

        cls.aAssetScrapers = cAttr.create("assetScrapers", "ascs")
        cAttr.array = True
        cAttr.hidden = False
        cAttr.writable = True
        cAttr.addChild(cls.aAssetScraperName)
        cAttr.addChild(cls.aAssetScraperActive)
        om.MPxNode.addAttribute(cls.aAssetScrapers)

        conductorRender.aExtraAssets = tAttr.create(
            "extraAssets", "eass", om.MFnData.kString
        )
        tAttr.array = True
        tAttr.hidden = False
        tAttr.writable = True
        tAttr.usedAsFilename = True
        om.MPxNode.addAttribute(conductorRender.aExtraAssets)

    @classmethod
    def make_environment_atts(cls):
        cAttr = om.MFnCompoundAttribute()
        tAttr = om.MFnTypedAttribute()
        nAttr = om.MFnNumericAttribute()
        cls.aExtraEnvironmentKey = tAttr.create(
            "extraEnvironmentKey", "eek", om.MFnData.kString
        )
        cls.aExtraEnvironmentValue = tAttr.create(
            "extraEnvironmentValue", "eev", om.MFnData.kString
        )
        cls.aExtraEnvironmentExclusive = nAttr.create(
            "extraEnvironmentExclusive", "eee", om.MFnNumericData.kBoolean, False
        )
        cls.aExtraEnvironment = cAttr.create("extraEnvironment", "een")

        cAttr.hidden = False
        cAttr.writable = True
        cAttr.array = True
        cAttr.addChild(cls.aExtraEnvironmentKey)
        cAttr.addChild(cls.aExtraEnvironmentValue)
        cAttr.addChild(cls.aExtraEnvironmentExclusive)
        om.MPxNode.addAttribute(cls.aExtraEnvironment)

    @classmethod
    def make_metadata_atts(cls):
        cAttr = om.MFnCompoundAttribute()
        tAttr = om.MFnTypedAttribute()
        cls.aMetadataKey = tAttr.create(
            "metadataKey", "mdk", om.MFnData.kString
        )
        cls.aMetadataValue = tAttr.create(
            "metadataValue", "mdv", om.MFnData.kString
        )

        cls.aMetadata = cAttr.create("metadata", "md")

        cAttr.hidden = False
        cAttr.writable = True
        cAttr.array = True
        cAttr.addChild(cls.aMetadataKey)
        cAttr.addChild(cls.aMetadataValue)
        om.MPxNode.addAttribute(cls.aMetadata)

    @classmethod
    def make_retries_atts(cls):
        nAttr = om.MFnNumericAttribute()

        cls.aRetriesWhenFailed = nAttr.create(
            "retriesWhenFailed", "rwf", om.MFnNumericData.kInt, 1
        )
        nAttr.writable = True
        nAttr.keyable = True
        nAttr.storable = True
        om.MPxNode.addAttribute(cls.aRetriesWhenFailed)

        cls.aRetriesWhenPreempted = nAttr.create(
            "retriesWhenPreempted", "rwp", om.MFnNumericData.kInt, 1
        )
        nAttr.writable = True
        nAttr.keyable = True
        nAttr.storable = True
        om.MPxNode.addAttribute(cls.aRetriesWhenPreempted)

    @classmethod
    def make_upload_flag_atts(cls):
        nAttr = om.MFnNumericAttribute()

        cls.aUseUploadDaemon = nAttr.create(
            "useUploadDaemon", "uud", om.MFnNumericData.kBoolean, False
        )
        nAttr.writable = True
        nAttr.keyable = True
        nAttr.storable = True
        om.MPxNode.addAttribute(cls.aUseUploadDaemon)

        conductorRender.aUploadOnly = nAttr.create(
            "uploadOnly", "upo", om.MFnNumericData.kBoolean, False
        )
        nAttr.writable = True
        nAttr.keyable = True
        nAttr.storable = True
        om.MPxNode.addAttribute(conductorRender.aUploadOnly)

    @classmethod
    def make_frames_atts(cls):
        tAttr = om.MFnTypedAttribute()
        nAttr = om.MFnNumericAttribute()
        uAttr = om.MFnUnitAttribute()

        cls.aStartFrame = uAttr.create(
            "startFrame", "stf", om.MFnUnitAttribute.kTime, 1
        )
        uAttr.writable = True
        uAttr.keyable = False
        uAttr.storable = True
        om.MPxNode.addAttribute(cls.aStartFrame)

        cls.aEndFrame = uAttr.create(
            "endFrame", "enf", om.MFnUnitAttribute.kTime, 10
        )
        uAttr.writable = True
        uAttr.keyable = False
        uAttr.storable = True
        om.MPxNode.addAttribute(cls.aEndFrame)

        cls.aByFrame = nAttr.create(
            "byFrame", "byf", om.MFnNumericData.kInt, 1
        )
        nAttr.writable = True
        nAttr.keyable = False
        nAttr.storable = True
        om.MPxNode.addAttribute(cls.aByFrame)

        cls.aAnimation = nAttr.create(
            "animation", "ani", om.MFnNumericData.kBoolean
        )
        nAttr.writable = True
        nAttr.keyable = False
        nAttr.storable = True
        om.MPxNode.addAttribute(cls.aAnimation)

        cls.aChunkSize = nAttr.create(
            "chunkSize", "csz", om.MFnNumericData.kInt, 1
        )
        nAttr.writable = True
        nAttr.keyable = False
        nAttr.storable = True
        om.MPxNode.addAttribute(cls.aChunkSize)

        cls.aUseCustomRange = nAttr.create(
            "useCustomRange", "ucr", om.MFnNumericData.kBoolean, False
        )
        nAttr.writable = True
        nAttr.keyable = False
        nAttr.storable = True
        om.MPxNode.addAttribute(cls.aUseCustomRange)

        cls.aCustomRange = tAttr.create(
            "customRange", "crn", om.MFnData.kString
        )
        tAttr.writable = True
        tAttr.storable = True
        om.MPxNode.addAttribute(cls.aCustomRange)

        cls.aUseScoutFrames = nAttr.create(
            "useScoutFrames", "usf", om.MFnNumericData.kBoolean, True
        )
        nAttr.writable = True
        nAttr.keyable = False
        nAttr.storable = True
        om.MPxNode.addAttribute(cls.aUseScoutFrames)

        cls.aScoutFrames = tAttr.create(
            "scoutFrames", "scf", om.MFnData.kString
        )
        tAttr.writable = True
        tAttr.storable = True
        om.MPxNode.addAttribute(cls.aScoutFrames)

        cls.aCurrentFrame = uAttr.create(
            "currentFrame", "cf", om.MFnUnitAttribute.kTime)
        uAttr.writable = True
        uAttr.keyable = False
        uAttr.storable = True
        om.MPxNode.addAttribute(cls.aCurrentFrame)

    @classmethod
    def make_task_atts(cls):
        tAttr = om.MFnTypedAttribute()
        cls.aTaskTemplate = tAttr.create(
            "taskTemplate", "ttm", om.MFnData.kString
        )
        tAttr.writable = True
        tAttr.storable = True
        om.MPxNode.addAttribute(cls.aTaskTemplate)

    @classmethod
    def make_notification_atts(cls):
        cAttr = om.MFnCompoundAttribute()
        tAttr = om.MFnTypedAttribute()
        nAttr = om.MFnNumericAttribute()

        cls.aEmailAddress = tAttr.create(
            "emailAddress", "eml", om.MFnData.kString
        )

        cls.aEmailAddressActive = nAttr.create(
            "emailAddressActive", "emla", om.MFnNumericData.kBoolean, True
        )

        cls.aEmailAddresses = cAttr.create("emailAddresses", "emls")

        cAttr.hidden = False
        cAttr.writable = True
        cAttr.array = True
        cAttr.addChild(cls.aEmailAddress)
        cAttr.addChild(cls.aEmailAddressActive)
        om.MPxNode.addAttribute(cls.aEmailAddresses)

    @classmethod
    def make_misc_atts(cls):
        tAttr = om.MFnTypedAttribute()
        cls.aDestinationDirectory = tAttr.create(
            "destinationDirectory", "ddr", om.MFnData.kString
        )
        tAttr.writable = True
        tAttr.storable = True
        tAttr.usedAsFilename = True
        om.MPxNode.addAttribute(cls.aDestinationDirectory)

        cls.aLocationTag = tAttr.create(
            "locationTag", "lct", om.MFnData.kString
        )
        tAttr.writable = True
        tAttr.storable = True
        om.MPxNode.addAttribute(cls.aLocationTag)

    @classmethod
    def make_developer_atts(cls):
        nAttr = om.MFnNumericAttribute()
        tAttr = om.MFnTypedAttribute()

        cls.aShowTracebacks = nAttr.create(
            "showTracebacks", "trc", om.MFnNumericData.kBoolean, False
        )
        nAttr.writable = True
        nAttr.storable = True
        nAttr.hidden = True
        om.MPxNode.addAttribute(cls.aShowTracebacks)

        cls.aUseFixtures = nAttr.create(
            "useFixtures", "ufx", om.MFnNumericData.kBoolean, False
        )
        nAttr.writable = True
        nAttr.storable = True
        # nAttr.hidden = True
        om.MPxNode.addAttribute(cls.aUseFixtures)


        cls.aFixturesDirectory = tAttr.create(
            "fixturesDirectory", "fdr", om.MFnData.kString
        )
        tAttr.writable = True
        tAttr.storable = True
        tAttr.usedAsFilename = True
        om.MPxNode.addAttribute(cls.aFixturesDirectory)



    @classmethod
    def make_autosave_atts(cls):

        tAttr = om.MFnTypedAttribute()
        nAttr = om.MFnNumericAttribute()

        cls.aAutosaveTemplate = tAttr.create(
            "autosaveTemplate", "ast", om.MFnData.kString
        )
        tAttr.writable = True
        tAttr.storable = True
        om.MPxNode.addAttribute(cls.aAutosaveTemplate)

        cls.aAutosave = nAttr.create(
            "autosave", "aus", om.MFnNumericData.kBoolean, True
        )
        nAttr.writable = True
        nAttr.storable = True
        nAttr.hidden = True
        om.MPxNode.addAttribute(cls.aAutosave)

        cls.aCleanupAutosave = nAttr.create(
            "cleanupAutosave", "cua", om.MFnNumericData.kBoolean, True
        )
        nAttr.writable = True
        nAttr.keyable = True
        nAttr.storable = True
        om.MPxNode.addAttribute(cls.aCleanupAutosave)

    @classmethod
    def make_hidden_atts(cls):
        nAttr = om.MFnNumericAttribute()
        cls.aDoScrape = nAttr.create(
            "doScrape", "dsc", om.MFnNumericData.kBoolean, False
        )
        nAttr.writable = True
        nAttr.storable = True
        nAttr.hidden = True
        om.MPxNode.addAttribute(cls.aDoScrape)

        cls.aTaskLimit = nAttr.create(
            "taskLimit", "tsl", om.MFnNumericData.kInt, 10
        )
        nAttr.writable = True
        nAttr.storable = True
        nAttr.hidden = True
        om.MPxNode.addAttribute(cls.aTaskLimit)

    @classmethod
    def make_output_att(cls):
        """
        Output atttribute.
        """
        tAttr = om.MFnTypedAttribute()
        cls.aOutput = tAttr.create("output", "out", om.MFnData.kString)
        tAttr.readable = True
        tAttr.storable = False
        tAttr.writable = False
        tAttr.keyable = False

        om.MPxNode.addAttribute(cls.aOutput)

    @classmethod
    def setup_attribute_affects(cls):
        om.MPxNode.attributeAffects(cls.aTitle, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aInstanceTypeName, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aProjectName, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aChunkSize, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aUseCustomRange, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aCustomRange, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aStartFrame, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aEndFrame, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aByFrame, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aAnimation, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aUseScoutFrames, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aScoutFrames, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aHostSoftware, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aPluginSoftware, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aExtraEnvironment, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aTaskTemplate, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aTaskLimit, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aDoScrape, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aExtraAssets, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aAssetScrapers, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aMetadata, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aRetriesWhenFailed, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aRetriesWhenPreempted, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aUseUploadDaemon, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aUploadOnly, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aEmailAddresses, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aDestinationDirectory, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aLocationTag, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aCurrentRenderLayer, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aCurrentRenderer, cls.aOutput)
        om.MPxNode.attributeAffects(cls.aCurrentFrame, cls.aOutput)

        om.MPxNode.attributeAffects(cls.aUseCustomRange, cls.aFrameCount)
        om.MPxNode.attributeAffects(cls.aCustomRange, cls.aFrameCount)
        om.MPxNode.attributeAffects(cls.aStartFrame, cls.aFrameCount)
        om.MPxNode.attributeAffects(cls.aEndFrame, cls.aFrameCount)
        om.MPxNode.attributeAffects(cls.aByFrame, cls.aFrameCount)
        om.MPxNode.attributeAffects(cls.aAnimation, cls.aFrameCount)

        om.MPxNode.attributeAffects(cls.aChunkSize, cls.aTaskCount)
        om.MPxNode.attributeAffects(cls.aUseCustomRange, cls.aTaskCount)
        om.MPxNode.attributeAffects(cls.aCustomRange, cls.aTaskCount)
        om.MPxNode.attributeAffects(cls.aStartFrame, cls.aTaskCount)
        om.MPxNode.attributeAffects(cls.aEndFrame, cls.aTaskCount)
        om.MPxNode.attributeAffects(cls.aByFrame, cls.aTaskCount)
        om.MPxNode.attributeAffects(cls.aAnimation, cls.aTaskCount)

        om.MPxNode.attributeAffects(cls.aUseScoutFrames, cls.aScoutTaskCount)
        om.MPxNode.attributeAffects(cls.aScoutFrames, cls.aScoutTaskCount)
        om.MPxNode.attributeAffects(cls.aChunkSize, cls.aScoutTaskCount)
        om.MPxNode.attributeAffects(cls.aUseCustomRange, cls.aScoutTaskCount)
        om.MPxNode.attributeAffects(cls.aCustomRange, cls.aScoutTaskCount)
        om.MPxNode.attributeAffects(cls.aStartFrame, cls.aScoutTaskCount)
        om.MPxNode.attributeAffects(cls.aEndFrame, cls.aScoutTaskCount)
        om.MPxNode.attributeAffects(cls.aByFrame, cls.aScoutTaskCount)
        om.MPxNode.attributeAffects(cls.aAnimation, cls.aScoutTaskCount)
        om.MPxNode.attributeAffects(cls.aCurrentFrame, cls.aScoutTaskCount)
  
        om.MPxNode.attributeAffects(cls.aUseCustomRange, cls.aFrameSpec)
        om.MPxNode.attributeAffects(cls.aCustomRange, cls.aFrameSpec)
        om.MPxNode.attributeAffects(cls.aStartFrame, cls.aFrameSpec)
        om.MPxNode.attributeAffects(cls.aEndFrame, cls.aFrameSpec)
        om.MPxNode.attributeAffects(cls.aByFrame, cls.aFrameSpec)
        om.MPxNode.attributeAffects(cls.aAnimation, cls.aFrameSpec)
        om.MPxNode.attributeAffects(cls.aCurrentFrame, cls.aFrameSpec)
   
        om.MPxNode.attributeAffects(cls.aUseScoutFrames, cls.aScoutSpec)
        om.MPxNode.attributeAffects(cls.aScoutFrames, cls.aScoutSpec)
        om.MPxNode.attributeAffects(cls.aChunkSize, cls.aScoutSpec)
        om.MPxNode.attributeAffects(cls.aUseCustomRange, cls.aScoutSpec)
        om.MPxNode.attributeAffects(cls.aCustomRange, cls.aScoutSpec)
        om.MPxNode.attributeAffects(cls.aStartFrame, cls.aScoutSpec)
        om.MPxNode.attributeAffects(cls.aEndFrame, cls.aScoutSpec)
        om.MPxNode.attributeAffects(cls.aByFrame, cls.aScoutSpec)
        om.MPxNode.attributeAffects(cls.aAnimation, cls.aScoutSpec)
        om.MPxNode.attributeAffects(cls.aCurrentFrame, cls.aScoutSpec)


    def compute(self, plug, data):
        """Compute output json from input attribs."""
        if (not ((plug == self.aOutput) or
                 (plug == self.aFrameCount) or
                 (plug == self.aTaskCount) or
                 (plug == self.aScoutTaskCount) or
                 (plug == self.aFrameSpec)or
                 (plug == self.aScoutSpec)
                 )):
            return None

        try:
            sequence = self.get_sequence(data)
        except (ValueError, TypeError):
            om.MGlobal.displayWarning("Invalid frame sequence specified.")
            return None

        scout_sequence = self.get_scout_sequence(data, sequence)
        frame_count = len(sequence)
        task_count = sequence.chunk_count()

        scout_task_count = 0
        scout_tasks_sequence = None
        if scout_sequence:
            scout_chunks = sequence.intersecting_chunks(scout_sequence)
            if scout_chunks:
                scout_tasks_sequence = Sequence.create(
                    ",".join(str(chunk) for chunk in scout_chunks))
                scout_task_count = len(scout_chunks)

        self.set_frame_info_plugs(
            data, frame_count, task_count, scout_task_count)

        self.set_frame_and_scout_spec(data, sequence, scout_tasks_sequence)

        if (plug != self.aOutput):
            return self

        context = self.get_context(data)
        expander = Expander(**context)

        handle = data.outputValue(self.aOutput)
        result = {}
        result.update(self.get_software_environment(data))
        result.update(self.get_instance_type(data))
        result.update(self.get_title(data, expander))
        result.update(self.get_project(data))
        result.update(self.get_tasks(data, sequence, context))
        result.update(self.get_upload_paths(data))
        result.update(self.get_upload_flags(data))
        result.update(self.get_retry_policy(data))
        result.update(self.get_notifiations(data))
        result.update(self.get_scout_frames(scout_sequence))
        result.update(self.get_metadata(data, expander))
        result.update(self.get_location_tag(data))
        result.update(self.get_destination_directory(data))

        handle.setString(json.dumps(result))

        data.setClean(plug)
        return self

    @classmethod
    def get_sequence(cls, data):
        chunk_size = data.inputValue(cls.aChunkSize).asInt()
        use_custom_range = data.inputValue(cls.aUseCustomRange).asBool()
        if use_custom_range:
            custom_range = data.inputValue(cls.aCustomRange).asString()
            return Sequence.create(custom_range, chunk_size=chunk_size, chunk_strategy="progressions")

        if not data.inputValue(cls.aAnimation).asBool():

            return Sequence.create(int(data.inputValue(cls.aCurrentFrame).asTime().asUnits(om.MTime.uiUnit())))

        start_frame = data.inputValue(
            cls.aStartFrame).asTime().asUnits(om.MTime.uiUnit())
        end_frame = data.inputValue(
            cls.aEndFrame).asTime().asUnits(om.MTime.uiUnit())
        by_frame = data.inputValue(cls.aByFrame).asInt()
        return Sequence.create(int(start_frame), int(end_frame), by_frame, chunk_size=chunk_size, chunk_strategy="progressions")

    @classmethod
    def get_scout_sequence(cls, data, main_sequence):
        use_scout_frames = data.inputValue(cls.aUseScoutFrames).asBool()
        if not use_scout_frames:
            return

        scout_frames = data.inputValue(cls.aScoutFrames).asString().strip()

        match = re.compile(r"^auto[, :]+(\d+)$").match(scout_frames)
        if match:
            samples = int(match.group(1))
            return main_sequence.subsample(samples)

        try:
            return Sequence.create(scout_frames)
        except (ValueError, TypeError):
            return

    @classmethod
    def get_scout_frames(cls, scout_sequence):
        return {"scout_frames": ",".join([str(s) for s in scout_sequence or []])}

    @classmethod
    def set_frame_info_plugs(cls, data, frame_count, task_count, scout_task_count):
        handle = data.outputValue(cls.aFrameCount)
        handle.setInt(frame_count)
        handle.setClean()
        handle = data.outputValue(cls.aTaskCount)
        handle.setInt(task_count)
        handle.setClean()
        handle = data.outputValue(cls.aScoutTaskCount)
        handle.setInt(scout_task_count)
        handle.setClean()

    @classmethod
    def set_frame_and_scout_spec(cls, data, sequence, scout_sequence):
        handleFrameSpec = data.outputValue(cls.aFrameSpec)
        handleScoutSpec = data.outputValue(cls.aScoutSpec)
        if scout_sequence:
            handleScoutSpec.setString(str(scout_sequence))
        else:
            handleScoutSpec.setString("")
        handleFrameSpec.setString(str(sequence))

        handleScoutSpec.setClean()
        handleFrameSpec.setClean()

    @classmethod
    def get_instance_type(cls, data):
        return {
            "instance_type": data.inputValue(cls.aInstanceTypeName).asString(),
            "preemptible": data.inputValue(cls.aPreemptible).asBool()
        }

    @classmethod
    def get_title(cls, data, expander):
        title = data.inputValue(cls.aTitle).asString()
        return {"job_title": expander.evaluate(title)}

    @classmethod
    def get_project(cls, data):
        return {"project": data.inputValue(cls.aProjectName).asString()}

    @classmethod
    def get_software_environment(cls, data):
        extra_env = cls.get_extra_env(data)
        packages_data = cls.get_software_packages(data)
        packages_data["env"].extend(extra_env)
        return {
            "environment": dict(packages_data["env"]),
            "software_package_ids": packages_data["ids"]
        }

    @classmethod
    def get_software_packages(cls, data):
        tree_data = coredata.data().get("software")
        paths = []
        host_path = data.inputValue(cls.aHostSoftware).asString()
        paths.append(host_path)
        array_handle = data.inputArrayValue(cls.aPluginSoftware)

        while not array_handle.isDone():
            plugin_path = "{}/{}".format(host_path,
                                         array_handle.inputValue().asString())
            paths.append(plugin_path)
            array_handle.next()

        result = {
            "ids": [],
            "env": PackageEnvironment()

        }

        for package in filter(None, [tree_data.find_by_path(path) for path in paths if path]):
            if package:
                result["ids"].append(package["package_id"])
                result["env"].extend(package)

        return result

    @classmethod
    def get_extra_env(cls, data):
        result = []
        array_handle = data.inputArrayValue(cls.aExtraEnvironment)
        while not array_handle.isDone():
            name = array_handle.inputValue().child(
                cls.aExtraEnvironmentKey).asString()
            value = array_handle.inputValue().child(
                cls.aExtraEnvironmentValue).asString()
            exclusive = array_handle.inputValue().child(
                cls.aExtraEnvironmentExclusive).asBool()
            name = name.strip()
            value = value.strip()

            if name and value:
                result.append(
                    {
                        "name": name,
                        "value": value,
                        "merge_policy": "exclusive" if exclusive else "append"
                    }
                )
            array_handle.next()
        return result

    def get_context(self, data):
        node = om.MFnDependencyNode(self.thisMObject())
        file_name = pm.sceneName()
        scene_name = os.path.splitext(os.path.split(file_name)[1])[0]

        context = {
            "Scene": "undefined",
            "SceneFile": "undefined",
            "Object": node.name(),
            "RenderLayer": layer_utils.get_current_layer_name(),
            "WorkspacePath": "undefined",
            "OutputPath":  "undefined",
            "Renderer": data.inputValue(self.aCurrentRenderer).asString(),
            "ConductorVersion": pm.moduleInfo(version=True, moduleName="conductor")
        }

        if file_name:
            context["Scene"] = scene_name
            context["SceneFile"] = Path(file_name).posix_path(with_drive=False)
        try:
            context["WorkspacePath"] = Path(pm.workspace(
                q=True, rd=True)).posix_path(with_drive=False)
        except:
            pass

        try:
            context["OutputPath"] = Path(data.inputValue(
                self.aDestinationDirectory).asString().strip()).posix_path(with_drive=False)
        except:
            pass

        low_context = {}
        for key in context:
            key_lower = key.lower()
            if not key_lower in context:
                low_context[key_lower] = context[key]
        context.update(low_context)
        return context

    @classmethod
    def get_tasks(cls, data, sequence, context):
        if data.inputValue(cls.aUploadOnly).asBool():
            return {}

        tasks = []
        template = data.inputValue(cls.aTaskTemplate).asString()
        limit = data.inputValue(cls.aTaskLimit).asInt()
        chunks = sequence.chunks()
        if limit < 0:
            limit = len(chunks)
        for i, chunk in enumerate(chunks):
            if i >= limit:
                break
            task_context = {
                "start": str(chunk.start),
                "end": str(chunk.end),
                "step": str(chunk.step),
                "chunk_length":  str(len(chunk))
            }
            task_context.update(context)
            expander = Expander(**task_context)

            tasks.append({
                "command": expander.evaluate(template),
                "frames": str(chunk)
            })
        return {"tasks_data": tasks}

    @classmethod
    def get_upload_flags(cls, data):
        return {
            "upload_only": data.inputValue(cls.aUploadOnly).asBool(),
            "local_upload": not data.inputValue(cls.aUseUploadDaemon).asBool()
        }

    @classmethod
    def get_retry_policy(cls, data):
        return {
            "autoretry_policy": {
                "preempted": {"max_retries":  data.inputValue(cls.aRetriesWhenPreempted).asInt()},
                "failed":    {"max_retries":   data.inputValue(cls.aRetriesWhenFailed).asInt()}
            }
        }

    @classmethod
    def get_notifiations(cls, data):
        result = []
        array_handle = data.inputArrayValue(cls.aEmailAddresses)
        while not array_handle.isDone():
            if array_handle.inputValue().child(
                    cls.aEmailAddressActive).asBool():
                value = array_handle.inputValue().child(
                    cls.aEmailAddress).asString().strip()
            if value:
                result.append(value)
            array_handle.next()
        return {"notify": result}

    def get_upload_paths(self, data):
        node_name = om.MFnDependencyNode(self.thisMObject()).name()

        path_list = PathList()
        path_list.add(*self.get_scraped_paths(data, node_name))
        path_list.add(*self.get_cached_paths(data))
        path_list.remove_missing()
        path_list.glob()
        return {"upload_paths": sorted([p.posix_path() for p in path_list])}

    @classmethod
    def get_scraped_paths(cls, data, node):
        result = []
        if not data.inputValue(cls.aDoScrape).asBool():
            return result
        array_handle = data.inputArrayValue(cls.aAssetScrapers)
        while not array_handle.isDone():
            if array_handle.inputValue().child(cls.aAssetScraperActive).asBool():
                script = array_handle.inputValue().child(cls.aAssetScraperName).asString()
                try:
                    scraper_module = importlib.import_module(script)
                    reload(scraper_module)
                    result += scraper_module.run(node)
                except SyntaxError:
                    om.MGlobal.displayWarning(
                        "Syntax error in the scraper script: '{}'.".format(script))
                    raise
                except ImportError:
                    om.MGlobal.displayWarning(
                        "Can't load the script '{}' as a Python module. Skipping!!".format(script))
                except BaseException:
                    om.MGlobal.displayWarning(
                        "Unknown problem with scraper: '{}'. Skipping!!".format(script))
                
            array_handle.next()

        return [p["path"] for p in result]

    @classmethod
    def get_cached_paths(cls, data):
        result = []
        array_handle = data.inputArrayValue(cls.aExtraAssets)
        while not array_handle.isDone():
            path = array_handle.inputValue().asString().strip()
            if path:
                result.append(path)
            array_handle.next()
        return result

    @classmethod
    def get_metadata(cls, data, expander):
        metadata = {}
        array_handle = data.inputArrayValue(cls.aMetadata)
        while not array_handle.isDone():
            key = array_handle.inputValue().child(
                cls.aMetadataKey).asString().strip()
            value = array_handle.inputValue().child(
                cls.aMetadataValue).asString().strip()

            metadata[key] = value

            array_handle.next()

        return {"metadata": expander.evaluate(metadata)}

    @classmethod
    def get_location_tag(cls, data):
        result = data.inputValue(cls.aLocationTag).asString().strip()
        return {"location": result} if result else {}

    @classmethod
    def get_destination_directory(cls, data):
        return {"output_path": data.inputValue(cls.aDestinationDirectory).asString().strip()}



from ciomaya.lib import software

"""
A scraper to collect paths from Yeti nodes.
"""
 

def run(_):
    ############# FOR NOW ###########
    if software.detect_yeti():
        raise NotImplementedError(
            "Yeti Asset Scraping is not yet implemented. To submit while the Yeti plugin is loaded, turn off the Yeti scraper in the Asset Scapers section. If you want to render yeti hair, we recommend you use the traditional client-tools submitter in the meantime.")
    return []
    #################################



import pymel.core as pm
import os
import glob
import webbrowser
from ciomaya.lib import window
from ciomaya.lib import software
from ciomaya.lib import const as k

MAYA_PARENT_WINDOW = 'MayaWindow'
CONDUCTOR_MENU = 'ConductorMenu'
CONDUCTOR_DOCS = 'https://docs.conductortech.com/'
LOG_LEVELS = ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"]
DEFAULT_LOG_LEVEL = LOG_LEVELS[2]


def unload():
    print "UNLOAD"

    if pm.menu(CONDUCTOR_MENU, q=True, exists=True):
        pm.menu(CONDUCTOR_MENU, e=True, deleteAllItems=True)
        pm.deleteUI(CONDUCTOR_MENU)


def load():
    print "LOAD"
    # unload()
    ConductorMenu()


class ConductorMenu(object):

    def __init__(self):
        if not pm.about(batch=True):
            pm.setParent(MAYA_PARENT_WINDOW)
            self.menu = pm.menu(CONDUCTOR_MENU, label="Conductor",
                                tearOff=True, pmc=pm.Callback(self.post_menu_command))
            self.jobs_menu = pm.menuItem(label="Submitter", subMenu=True)

            pm.setParent(self.menu, menu=True)

            pm.menuItem(divider=True)

            self.log_level_menu = self.build_log_level_menu()
            pm.setParent(self.menu, menu=True)

            pm.menuItem(divider=True)

            self.help_menu = pm.menuItem(label="Help", command=pm.Callback(
                webbrowser.open, CONDUCTOR_DOCS, new=2))
            self.about_menu = pm.menuItem(
                label="About", command=pm.Callback(window.show_about))

    def build_log_level_menu(self):
        result = pm.menuItem(label="Log level", subMenu=True)

        for level in LOG_LEVELS:
            pm.menuItem(label=level, radioButton=(level == DEFAULT_LOG_LEVEL),
                        command=pm.Callback(self.set_log_level, level))
        return result

    def post_menu_command(self):
        """
        Build the Select/Create submenu just before the menu is opened.
        """
        pm.setParent(self.jobs_menu,   menu=True)
        pm.menu(self.jobs_menu, edit=True, deleteAllItems=True)
        for j in pm.ls(type="conductorRender"):

            pm.menuItem(label="Select {}".format(str(j)),
                        command=pm.Callback(select_and_show, j))
        pm.menuItem(divider=True)

        pm.menuItem(label="Create",  command=pm.Callback(create_render_node))
        pm.setParent(self.menu, menu=True)

    def set_log_level(self, level):
        print "Setting Conductor log level to '{}'".format(level)


def create_render_node():
    node = pm.createNode("conductorRender")

    setup_render_node(node)
    select_and_show(node)


def select_and_show(node):
    pm.select(node)

    if not pm.mel.isAttributeEditorRaised():
        pm.mel.openAEWindow()


def setup_render_node(node):
    dest_path = os.path.join(
        pm.workspace(q=True, rd=True),

        pm.workspace.fileRules.get("images")
    )

    node.attr("destinationDirectory").set(dest_path)
    node.attr("hostSoftware").set(software.detect_host())


    ##### WE MUST DRY THIS UP ()TODAY! Why? Read John Ousterhout. #######
    # Combine with stuff ins AEsoftware Also - make each detection routine
    # discoverable through introspection, perhapes by making a SoftwareDetect()
    # class and deriving from it? 
    i = 0
    for plugin_software_name in filter(None, [
        software.detect_mtoa(),
        software.detect_rfm(),
        software.detect_vray(),
        software.detect_redshift(),
        software.detect_yeti(),
        
    ]):
        node.attr("pluginSoftware")[i].set(plugin_software_name)
        i += 1

    node.attr("taskTemplate").set(k.DEFAULT_TEMPLATE)
    node.attr("autosaveTemplate").set(k.DEFAULT_AUTOSAVE_TEMPLATE)
    node.attr("instanceTypeName").set(k.DEFAULT_INSTANCE_TYPE)
    node.attr("title").set(k.DEFAULT_TITLE)

    node.attr("customRange").set("1-10")
    node.attr("scoutFrames").set("auto:3")

    pm.Attribute("defaultRenderGlobals.startFrame") >> node.attr("startFrame")
    pm.Attribute("defaultRenderGlobals.endFrame") >> node.attr("endFrame")
    pm.Attribute("defaultRenderGlobals.byFrameStep") >> node.attr("byFrame")
    pm.Attribute("defaultRenderGlobals.animation") >> node.attr("animation")
    pm.Attribute("defaultRenderGlobals.currentRenderer") >> node.attr(
        "currentRenderer")
    pm.Attribute("renderLayerManager.currentRenderLayer") >> node.attr(
        "currentRenderLayer")
    pm.Attribute("time1.outTime") >> node.attr("currentFrame")

    # asset scrapers
    script_path = os.path.join(pm.moduleInfo(
        path=True, moduleName="conductor"), "scripts")

    files = sorted(glob.glob("{}/scrape_*.py".format(script_path)))
    for i, scraper in enumerate(files):
        scraper_name = os.path.splitext(os.path.basename(scraper))[0]
        node.attr("assetScrapers[{:d}].assetScraperName".format(i)).set(
            scraper_name)



"""
Handle the UI for extra assets:
"""

import pymel.core as pm
 
from ciomaya.lib.ae import AEcommon


def create_ui(node_attr):
    """Build static UI.
    """
 
    with AEcommon.ae_template():

        form = pm.formLayout(nd=100, height=200)
        clear_all_btn = pm.button("extraAssetsClearAllButton", label="Clear All", height=24)
        clear_sel_btn = pm.button("extraAssetsClearSelButton",label="Remove Selection", height=24)
        browse_file_btn = pm.button("extraAssetsBrowseFileButton",label="Browse Files", height=24)
        browse_dir_btn = pm.button("extraAssetsBrowseDirButton",label="Browse Folder", height=24)
        scroll_list = pm.textScrollList("extraAssetsScrollList",numberOfRows=10, allowMultiSelection=True)

        form.attachForm(clear_all_btn, "left", 2)
        form.attachPosition(clear_all_btn, "right", 2, 25)
        form.attachForm(clear_all_btn, "top", 2)
        form.attachNone(clear_all_btn, "bottom")

        form.attachPosition(clear_sel_btn, "left", 2, 25)
        form.attachPosition(clear_sel_btn, "right", 2, 50)
        form.attachForm(clear_sel_btn, "top", 2)
        form.attachNone(clear_sel_btn, "bottom")

        form.attachPosition(browse_file_btn, "left", 2, 50)
        form.attachPosition(browse_file_btn, "right", 2, 75)
        form.attachForm(browse_file_btn, "top", 2)
        form.attachNone(browse_file_btn, "bottom")

        form.attachPosition(browse_dir_btn, "left", 2, 75)
        form.attachForm(browse_dir_btn, "right", 2)
        form.attachForm(browse_dir_btn, "top", 2)
        form.attachNone(browse_dir_btn, "bottom")

        form.attachForm(scroll_list, "left", 2)
        form.attachForm(scroll_list, "right", 2)
        form.attachControl(scroll_list, "top", 2, clear_sel_btn)
        form.attachForm(scroll_list, "bottom", 2)

        pm.textScrollList(scroll_list, edit=True, append=("1", "2", "3"))

        pm.setParent("..")
     
    
        populate_ui(node_attr)


def populate_ui(node_attr):
    """Reconfigure action buttons when node changes."""
    widgets = _get_widgets()
    attr = pm.Attribute(node_attr)

    pm.textScrollList(widgets["scroll_list"], edit=True, removeAll=True)
    paths = filter(None, [asset_attr.get() for asset_attr in pm.Attribute(node_attr)])
    pm.textScrollList(widgets["scroll_list"], edit=True, append=paths)

    pm.button(
        widgets["clear_all_btn"],
        edit=True,
        command=pm.Callback(_on_clear_all_btn, attr)
    )
    pm.button(
        widgets["clear_sel_btn"],
        edit=True,
        command=pm.Callback(_on_clear_sel_btn, attr,  widgets["scroll_list"]),
    )
    pm.button(
        widgets["browse_file_btn"],
        edit=True,
        command=pm.Callback(_on_browse_btn, attr, 4),
    )
    pm.button(
        widgets["browse_dir_btn"],
        edit=True,
        command=pm.Callback(_on_browse_btn, attr, 3),
    )



def _get_widgets(parent=None):
    if not parent:
        parent = pm.setParent(q=True)
    return {
        "clear_all_btn":  AEcommon.find_ui("extraAssetsClearAllButton", parent),
        "clear_sel_btn":  AEcommon.find_ui("extraAssetsClearSelButton", parent),
        "browse_file_btn":  AEcommon.find_ui("extraAssetsBrowseFileButton", parent),
        "browse_dir_btn":  AEcommon.find_ui("extraAssetsBrowseDirButton", parent),
        "scroll_list":  AEcommon.find_ui("extraAssetsScrollList", parent)
    }

 
def _on_clear_all_btn(attr):
    for element in attr:
        pm.removeMultiInstance(element, b=True)

def _on_clear_sel_btn(attr, scroll_list):
    sel_indices = [
        i - 1
        for i in pm.textScrollList(  scroll_list, q=True, selectIndexedItem=True
        )
    ]
    logical_indices = attr.getArrayIndices()
    for i in sel_indices:
        pm.removeMultiInstance(attr[logical_indices[i]], b=True)

def _on_browse_btn(attr, mode):
    caption = "Choose Files" if mode == 4 else "Choose Folder"

    entries = pm.fileDialog2(
        caption=caption,
        okCaption="Choose",
        fileFilter="*",
        dialogStyle=2,
        fileMode=mode,
        dir=pm.workspace.getPath(),
    )
    if entries:
        logical_indices = attr.getArrayIndices()
        next_index = logical_indices[-1] + 1 if logical_indices else 0
        for entry in entries:
            attr[next_index].set(entry)
            next_index += 1
        return
    pm.displayWarning("No files Selected")